/*
 * Copyright (C) 2026
 *
 * This file is part of libass.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// WARNING - THIS FILE IS AUTO-GENERATED. DO NOT EDIT IT MANUALLY.
// Regenerate with: python3 gen_arabic_charmap.py

#ifndef LIBASS_ASS_ARABIC_CHARMAP_H
#define LIBASS_ASS_ARABIC_CHARMAP_H

#include <stdint.h>

uint32_t ass_font_charmap_arabic_simplified(uint32_t symbol);
uint32_t ass_font_charmap_arabic_traditional(uint32_t symbol);

#endif /* LIBASS_ASS_ARABIC_CHARMAP_H */
